const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");

const Device = @import("device.zig").Device;
const ProfileMark = @import("../../common/profiler.zig").ProfileMark;
const settings = @import("settings.zig");
const Swapchain = @import("swapchain.zig").Swapchain;
const descriptor = @import("descriptor.zig");
const Buffer = @import("Buffer.zig");

var s_layout = vk.DescriptorSetLayout.null_handle;
var s_pool = vk.DescriptorPool.null_handle;
var s_sets = std.mem.zeroes([settings.resource_sets]vk.DescriptorSet);
var s_writes: [settings.Bid.Count]vk.WriteDescriptorSet = undefined;
var s_bindings: [settings.Bid.Count]Binding = undefined;
var s_dirty = std.mem.zeroes([settings.Bid.Count]i32);

const bindings = [_]settings.Bid{
    settings.Bid.TextureTable.T1D,
    settings.Bid.TextureTable.T2D,
    settings.Bid.TextureTable.T3D,
    settings.Bid.TextureTable.TCube,
    settings.Bid.TextureTable.T2DArray,
    settings.Bid.Globals,
    settings.Bid.SceneLuminance,
    settings.Bid.RWSceneLuminance,
    settings.Bid.HistogramBuffer,
    settings.Bid.ExposureBuffer,
};

pub const set_layout_bindings = LayoutBindings();
fn LayoutBindings() [bindings.len]vk.DescriptorSetLayoutBinding {
    comptime var bs: [bindings.len]vk.DescriptorSetLayoutBinding = undefined;
    assert(bindings.len == settings.Bid.Count);
    inline for (bs) |_, i| {
        bs[i] = .{
            .binding = bindings[i].id,
            .descriptor_type = bindings[i].desc_type,
            .descriptor_count = bindings[i].count,
            .stage_flags = bindings[i].stages,
            .p_immutable_samplers = null,
        };
    }

    return bs;
}

const pm_init = ProfileMark.init("Bindings.init");
pub fn init(device: *const Device) !void {
    pm_init.begin();
    defer pm_init.end();

    errdefer deinit(device);

    s_layout = try descriptor.Set.Layout.init(device, &set_layout_bindings, .{});
    s_pool = try descriptor.Pool.init(device, s_sets.len, &descriptor.Pool.sizes);

    for (s_sets) |*set| {
        set.* = try descriptor.Set.init(device, s_pool, s_layout);
    }
}

pub fn deinit(device: *const Device) void {
    if (s_pool != .null_handle) {
        for (s_sets) |*set| {
            descriptor.Set.deinit(device, s_pool, set.*);
            set.* = .null_handle;
        }

        descriptor.Pool.deinit(device, s_pool);
        s_pool = .null_handle;
    }

    descriptor.Set.Layout.deinit(device, s_layout);
    s_layout = .null_handle;
}

const pm_update = ProfileMark.init("Bindings.update");
pub fn update(device: *const Device, swapchain: *const Swapchain) void {
    pm_update.begin();
    defer pm_update.end();

    // kara says: why do validation layers think there is a command buffer in flight using this desc set?
    // submit.awaitAll();

    //const set = getSet(swapchain);
    // TexTable.write(set);
    flush(device, swapchain);
}

pub fn getSet(swapchain: *const Swapchain) vk.DescriptorSet {
    const sync_index = swapchain.sync_index;
    assert(sync_index < s_sets.len);
    assert(s_sets[sync_index] != .null_handle);
    return s_sets[sync_index];
}

pub fn getSetLayout() vk.DescriptorSetLayout {
    assert(s_layout != .null_handle);
    return s_layout;
}

const pm_flush = ProfileMark.init("Bindings.flush");
pub fn flush(device: *const Device, swapchain: *const Swapchain) void {
    pm_flush.begin();
    defer pm_flush.end();

    const set = getSet(swapchain);
    var write_count: usize = 0;
    for (s_bindings) |binding, i| {
        assert(s_dirty[i] >= 0);
        if (s_dirty[i] <= 0) {
            continue;
        }

        s_dirty[i] -= 1;

        var write = &s_writes[write_count];
        write_count += 1;
        write.* = .{
            .s_type = .write_descriptor_set,
            .descriptor_type = binding.desc_type,
            .dst_set = set,
            .dst_binding = @intCast(u32, i),
            .dst_array_element = 0,
            .descriptor_count = 1,
            .p_image_info = undefined,
            .p_buffer_info = undefined,
            .p_texel_buffer_view = undefined,
        };

        switch (binding.info) {
            .buffer => |*buffer| {
                switch (binding.desc_type) {
                    .uniform_buffer, .storage_buffer, .uniform_buffer_dynamic, .storage_buffer_dynamic => {
                        write.p_buffer_info = @ptrCast([*]const vk.DescriptorBufferInfo, buffer);
                    },
                    else => {
                        assert(false);
                    },
                }
            },
            .image => |*image| {
                switch (binding.desc_type) {
                    .sampler, .combined_image_sampler, .sampled_image, .storage_image, .input_attachment => {
                        write.p_image_info = @ptrCast([*]const vk.DescriptorImageInfo, image);
                    },
                    else => {
                        assert(false);
                    },
                }
            },
        }

        if (write_count > 0) {
            device.dispatch.updateDescriptorSets(device.handle, @intCast(u32, write_count), &s_writes, 0, undefined);
        }
    }
}

pub fn bindImage(id: usize, desc_type: vk.DescriptorType, sampler: vk.Sampler, view: vk.ImageView, layout: vk.ImageLayout) void {
    assert(id < s_bindings.len);
    assert(view != .null_handle and sampler != .null_handle);
    s_bindings[id].desc_type = desc_type;
    s_bindings[id].image.image_layout = layout;
    s_bindings[id].image.image_view = view;
    s_bindings[id].image.sampler = sampler;
    s_dirty[id] = s_sets.len;
}

pub fn bindBuffer(id: usize, desc_type: vk.DescriptorType, buffer: *const Buffer) void {
    assert(id < s_bindings.len);
    assert(buffer.handle != .null_handle);
    s_bindings[id].desc_type = desc_type;
    s_bindings[id].info.buffer.buffer = buffer.handle;
    s_bindings[id].info.buffer.offset = 0;
    s_bindings[id].info.buffer.range = buffer.size;
    s_dirty[id] = s_sets.len;
}

pub const Binding = struct {
    const Type = enum {
        buffer,
        image,
    };
    const Union = union(Type) {
        buffer: vk.DescriptorBufferInfo,
        image: vk.DescriptorImageInfo,
    };

    desc_type: vk.DescriptorType,
    info: Union,
};
