const std = @import("std");
const assert = @import("std").debug.assert;
const vk = @import("vulkan");

const Vec4f = @import("../../math/vec.zig").Vec4f;

const settings = @import("settings.zig");
const Pass = @import("Pass.zig");
const Swapchain = @import("swapchain.zig").Swapchain;
const Image = @import("image.zig").Image;
const Context = @import("Context.zig");
const Device = @import("device.zig").Device;
const DeviceDispatch = @import("device.zig").DeviceDispatch;
const ProfileMark = @import("../../common/profiler.zig").ProfileMark;
const Renderer = @import("Renderer.zig");
const SubmitId = @import("submit_id.zig").SubmitId;
const Fence = @import("sync.zig").Fence;
const Semaphore = @import("sync.zig").Semaphore;
const Bindings = @import("Bindings.zig");
const BufferBuffer = @import("Buffer.zig");
const meshes = @import("meshes.zig");

const queues = @import("queues.zig");

pub const write_access: vk.AccessFlags = .{
    .shader_write_bit = true,
    .color_attachment_write_bit = true,
    .depth_stencil_attachment_write_bit = true,
    .transfer_write_bit = true,
    .host_write_bit = true,
    .memory_write_bit = true,
    .transform_feedback_write_bit_ext = true,
    .transform_feedback_counter_write_bit_ext = true,
    .acceleration_structure_write_bit_khr = true,
    .command_preprocess_write_bit_nv = true,
};
pub const read_access: vk.AccessFlags = .{
    .indirect_command_read_bit = true,
    .index_read_bit = true,
    .vertex_attribute_read_bit = true,
    .uniform_read_bit = true,
    .input_attachment_read_bit = true,
    .shader_read_bit = true,
    .color_attachment_read_bit = true,
    .depth_stencil_attachment_read_bit = true,
    .transfer_read_bit = true,
    .host_read_bit = true,
    .memory_read_bit = true,
    .transform_feedback_counter_read_bit_ext = true,
    .conditional_rendering_read_bit_ext = true,
    .color_attachment_read_noncoherent_bit_ext = true,
    .acceleration_structure_read_bit_khr = true,
    .fragment_density_map_read_bit_ext = true,
    .fragment_shading_rate_attachment_read_bit_khr = true,
    .command_preprocess_read_bit_nv = true,
};

pub const CommandInitError = Pool.InitError || Fence.InitError || DeviceDispatch.AllocateCommandBuffersError;
const pm_init = ProfileMark.init("Command.init");
pub fn init(device: *const Device, queue: *queues.Queue, id: queues.QueueId) CommandInitError!void {
    pm_init.begin();
    defer pm_init.end();

    queue.index = id;
    queue.cmd_pool = try Pool.init(device, queue.family, .{ .reset_command_buffer_bit = true });
    errdefer queue.cmd_pool.deinit(device);

    try device.dispatch.allocateCommandBuffers(device.handle, &.{
        .command_pool = queue.cmd_pool.handle,
        .level = .primary,
        .command_buffer_count = settings.cmds_per_queue,
    }, &queue.cmds);

    for (queue.cmds) |_, i| {
        queue.cmd_fences[i] = try Fence.init(device, true);
        queue.cmd_ids[i] = 0;
    }
}

const pm_deinit = ProfileMark.init("Command.deinit");
pub fn deinit(device: *const Device, queue: *const queues.Queue) void {
    pm_init.begin();
    defer pm_init.end();

    for (queue.cmd_fences) |fence| {
        fence.deinit(device);
    }

    queue.cmd_pool.deinit(device);
}

pub fn getHeadSubmit(queue_id: queues.QueueId) SubmitId {
    const ctx = &Context.context;
    assert(@enumToInt(queue_id) < ctx.cur_cmd_buf.len);
    const cur = &ctx.cur_cmd_buf[@enumToInt(queue_id)];
    const prev = &ctx.prev_cmd_buf[@enumToInt(queue_id)];

    const cmd = if (cur.handle != .null_handle) cur else prev;
    var id = SubmitId{};
    if (cur.handle != .null_handle and cur.began) {
        id.counter = cmd.id;
        id.queue_id = cmd.queue_id;
        id.valid = cmd.handle != .null_handle;
    }
    return id;
}

const pm_flush = ProfileMark.init("Command.flush");
pub fn flush(device: *const Device) !void {
    pm_flush.begin();
    defer pm_flush.end();

    var ctx = &Context.context;
    for (ctx.cur_cmd_buf) |*buf| {
        if (buf.began) {
            _ = try buf.submit(device, null, .{}, null);
        }
    }
}

pub const Buffer = struct {
    const Self = @This();

    handle: vk.CommandBuffer = .null_handle,
    fence: Fence,
    id: u32,
    queue_id: queues.QueueId,
    // indicates which types of cmds are legal in this cmdbuf
    gfx: bool,
    comp: bool,
    xfer: bool,
    pres: bool,
    // cmd state
    began: bool,
    ended: bool,
    submitted: bool,
    in_render_pass: bool,
    subpass: u8,
    queue_transfer_src: bool,
    queue_transfer_dst: bool,

    const GetError = Buffer.InitError || DeviceDispatch.BeginCommandBufferError;
    pub fn get(queue_id: queues.QueueId, device: *const Device) GetError!*Buffer {
        var ctx = &Context.context;
        var cmd = &ctx.cur_cmd_buf[@enumToInt(queue_id)];
        if (cmd.handle == .null_handle) {
            try cmd.init(device, queues.get(queue_id));
        }

        if (!cmd.began) {
            try device.dispatch.beginCommandBuffer(cmd.handle, &.{
                .flags = .{ .one_time_submit_bit = true },
                .p_inheritance_info = null,
            });
            cmd.began = true;
            ctx.most_recent_begin = queue_id;
        }

        return cmd;
    }

    pub const InitError = Fence.WaitError || Fence.ResetError;
    const pm_buffer_init = ProfileMark.init("Command.Buffer.init");
    // XXX: I do not really understand this?
    pub fn init(self: *Self, device: *const Device, queue: *queues.Queue) InitError!void {
        pm_buffer_init.begin();
        defer pm_buffer_init.end();

        const ring_mask: u32 = queue.cmds.len - 1;
        // TODO: decode meaning of SASSERT((NELEM(queue->cmds) & (NELEM(queue->cmds) - 1u)) == 0);

        const head_id = queue.head;
        queue.head += 1;

        while (head_id - queue.tail >= queue.cmds.len) : (queue.tail +%= 1) {
            const tail_slot = queue.tail & ring_mask;
            assert(queue.cmd_ids[tail_slot] == queue.tail);
            const tail_fence = queue.cmd_fences[tail_slot];
            try tail_fence.wait(device);
            queue.cmd_ids[tail_slot] = 0;
        }

        const head_slot = head_id & ring_mask;
        assert(queue.cmd_ids[head_slot] == 0);
        queue.cmd_ids[head_slot] = head_id;
        const cmd = queue.cmds[head_slot];
        const fence = queue.cmd_fences[head_slot];

        assert((try fence.stat(device)) == .signaled);
        try fence.reset(device);

        self.handle = cmd;
        self.fence = fence;
        self.id = head_id;
        self.queue_id = queue.queue_id;
        self.gfx = queue.gfx;
        self.comp = queue.comp;
        self.xfer = queue.xfer;
        self.pres = queue.pres;
    }

    const pm_buffer_reset = ProfileMark.init("Command.Buffer.reset");
    pub fn reset(self: *const Self, device: *const Device) !void {
        pm_buffer_reset.begin();
        defer pm_buffer_reset.end();

        assert(!self.submitted);
        if (self.began and !self.ended) {
            try device.dispatch.endCommandBuffer(self.handle);
            self.ended = true;
        }

        try device.dispatch.resetCommandBuffer(self.handle, null);
        self.began = false;
        self.ended = false;
    }

    pub fn bindDescSets(self: *Self, device: *const Device, bindpoint: vk.PipelineBindPoint, layout: vk.PipelineLayout, set_count: u32, sets: [*]const vk.DescriptorSet) void {
        assert(self.handle != .null_handle);
        assert(self.gfx or self.comp);
        assert(layout != .null_handle);
        if (set_count > 0) {
            device.dispatch.cmdBindDescriptorSets(self.handle, bindpoint, layout, 0, set_count, sets, 0, undefined);
        }
    }

    pub fn bindPass(self: *Self, device: *const Device, swapchain: *const Swapchain, pass: *const Pass) void {
        assert(self.handle != .null_handle);
        assert(self.gfx or self.comp);
        assert(pass.pipeline != .null_handle);
        device.dispatch.cmdBindPipeline(self.handle, pass.bind_point, pass.pipeline);
        const set = Bindings.getSet(swapchain);
        self.bindDescSets(device, pass.bind_point, pass.layout, 1, @ptrCast([*]const vk.DescriptorSet, &set));
    }

    const pm_buffer_beginrenderpass = ProfileMark.init("Command.Buffer.beginRenderPass");
    pub fn beginRenderPass(self: *Self, device: *const Device, pass: vk.RenderPass, framebuf: vk.Framebuffer, rect: vk.Rect2D, clear_values: []const vk.ClearValue) void {
        pm_buffer_beginrenderpass.begin();
        defer pm_buffer_beginrenderpass.end();

        assert(self.handle != .null_handle);
        assert(!self.in_render_pass);
        device.dispatch.cmdBeginRenderPass(self.handle, &.{
            .render_pass = pass,
            .framebuffer = framebuf,
            .render_area = rect,
            .clear_value_count = @intCast(u32, clear_values.len),
            .p_clear_values = clear_values.ptr,
        }, .@"inline");
        self.in_render_pass = true;
        self.subpass = 0;
    }

    pub fn nextSubpass(self: *Self, device: *const Device, contents: vk.SubpassContents) void {
        assert(self.handle != .null_handle);
        assert(self.in_render_pass);
        device.dispatch.cmdNextSubpass(self.handle, contents);
        self.subpass += 1;
    }

    pub fn endRenderPass(self: *Self, device: *const Device) void {
        assert(self.handle != .null_handle);
        assert(self.in_render_pass);
        device.dispatch.cmdEndRenderPass(self.handle);
        self.in_render_pass = false;
        self.subpass = 0;
    }

    pub fn pushConstants(self: *Self, device: *const Device, pass: *const Pass, constants: anytype) void {
        const Ptr = @TypeOf(constants);

        comptime if (@typeInfo(Ptr) != .Pointer) {
            @compileError("constants argument provided to Command.pushConstants must be a pointer type, got " ++ @typeName(Ptr));
        };

        comptime if (@typeInfo(Ptr).Pointer.size != .One) {
            @compileError("constants argument provided to Command.pushConstants must be a single-item pointer");
        };

        const size = @sizeOf(@typeInfo(Ptr).Pointer.child);
        comptime if (size <= 0) {
            @compileError("constants argument provided to Command.pushConstants must have > 0 bytes.");
        };

        assert(self.handle != .null_handle);
        assert(self.gfx or self.comp);
        assert(size == pass.push_constant_bytes);
        assert(pass.layout != .null_handle);
        assert(pass.stage_flags.toInt() != 0);
        device.dispatch.cmdPushConstants(self.handle, pass.layout, pass.stage_flags, 0, size, constants);
    }

    pub fn viewport(self: *Self, device: *const Device, vp: vk.Viewport, scissor: vk.Rect2D) void {
        assert(self.handle != .null_handle);
        device.dispatch.cmdSetViewport(self.handle, 0, 1, @ptrCast([*]const vk.Viewport, &vp));
        device.dispatch.cmdSetScissor(self.handle, 0, 1, @ptrCast([*]const vk.Rect2D, &scissor));
    }

    pub fn defaultViewport(self: *Self, device: *const Device, swapchain: *const Swapchain) void {
        self.viewport(device, swapchain.getViewport(), swapchain.getRect());
    }

    pub fn draw(self: *Self, device: *const Device, vertex_count: i32, first_vertex: i32) void {
        assert(self.handle != .null_handle);
        assert(vertex_count > 0);
        assert(first_vertex >= 0);
        assert(self.in_render_pass);
        device.dispatch.cmdDraw(self.handle, vertex_count, 1, first_vertex, 0);
    }

    pub fn drawMesh(self: *Self, device: *const Device, mesh_id: meshes.Mesh.Id) void {
        assert(self.handle != .null_handle);
        assert(self.gfx);
        if (meshes.get(mesh_id)) |vk_mesh| {
            assert(vk_mesh.vert_count > 0);
            assert(vk_mesh.buffer.handle != .null_handle);

            if (vk_mesh.vert_count > 0) {
                const stream_size = vk_mesh.vert_count * @sizeOf(Vec4f);
                const buffers = [_]vk.Buffer{
                    vk_mesh.buffer.handle,
                    vk_mesh.buffer.handle,
                    vk_mesh.buffer.handle,
                    vk_mesh.buffer.handle,
                };
                const offsets = [_]vk.DeviceSize{
                    stream_size * 0,
                    stream_size * 1,
                    stream_size * 2,
                    stream_size * 3,
                };

                device.dispatch.cmdBindVertexBuffers(self.handle, 0, buffers.len, &buffers, &offsets);
                device.dispatch.cmdDraw(self.handle, vk_mesh.vert_count, 1, 0, 0);
            }
        }
    }

    pub fn touchBuffer(self: *Self, buf: *BufferBuffer) void {
        assert(self.handle != .null_handle);
        assert(buf.handle != .null_handle);
        buf.state.cmd_id = self.id;
    }

    pub fn touchImage(self: *Self, img: *Image) void {
        assert(self.handle != .null_handle);
        assert(img.handle != .null_handle);
        img.state.cmd_id = self.id;
    }

    const SubmitError = FlushQueueTransfersError || DeviceDispatch.EndCommandBufferError || DeviceDispatch.QueueSubmitError;
    const pm_buffer_submit = ProfileMark.init("Command.Buffer.submit");
    pub fn submit(self: *Self, device: *const Device, wait_sema: ?Semaphore, wait_mask: vk.PipelineStageFlags, signal_sema: ?Semaphore) SubmitError!SubmitId {
        pm_buffer_submit.begin();
        defer pm_buffer_submit.end();

        assert(self.began);
        assert(!self.ended);
        assert(!self.submitted);
        assert(!self.in_render_pass);

        const queue_id = self.queue_id;

        const queue = queues.get(queue_id);
        var submit_id = SubmitId{
            .counter = self.id,
            .queue_id = queue_id,
        };

        var ctx = &Context.context;

        const slot = self.id & (queue.cmds.len - 1);
        const fence = queue.cmd_fences[slot];

        if (self.queue_transfer_dst) {
            assert(!self.queue_transfer_src);
            try flushQueueTransfers(device);
        }

        assert(!self.queue_transfer_dst);
        assert(!self.ended);

        try device.dispatch.endCommandBuffer(self.handle);
        self.ended = true;

        assert((try fence.stat(device)) == .unsignaled);

        const submit_info = vk.SubmitInfo{
            .wait_semaphore_count = if (wait_sema != null) @as(u32, 1) else 0,
            .p_wait_semaphores = if (wait_sema) |s| @ptrCast([*]const vk.Semaphore, &s.handle) else undefined,
            .p_wait_dst_stage_mask = @ptrCast([*]const vk.PipelineStageFlags, &wait_mask),
            .signal_semaphore_count = if (signal_sema != null) @as(u32, 1) else 0,
            .p_signal_semaphores = if (signal_sema) |s| @ptrCast([*]const vk.Semaphore, &s.handle) else undefined,
            .command_buffer_count = 1,
            .p_command_buffers = @ptrCast([*]const vk.CommandBuffer, &self.handle),
        };

        try device.dispatch.queueSubmit(queue.handle, 1, @ptrCast([*]const vk.SubmitInfo, &submit_info), fence.handle);
        self.submitted = true;
        ctx.last_submit_queue = self.queue_id;

        // copy to prevCmd for debug purposes
        var prev_cmd = &ctx.prev_cmd_buf[@enumToInt(queue_id)];
        prev_cmd.* = self.*;

        // reset to blank slate
        self.* = std.mem.zeroes(Self);
        self.queue_id = queue_id;
        submit_id.valid = true;
        return submit_id;
    }

    pub const FlushQueueTransfersError = DeviceDispatch.EndCommandBufferError || DeviceDispatch.QueueSubmitError;
    const pm_buffer_flushqueuetransfers = ProfileMark.init("Command.Buffer.flushqueuetransfers");
    pub fn flushQueueTransfers(device: *const Device) FlushQueueTransfersError!void {
        pm_buffer_flushqueuetransfers.begin();
        defer pm_buffer_flushqueuetransfers.end();

        var ctx = Context.context;
        var submits: [ctx.cur_cmd_buf.len]bool = undefined;
        for (ctx.cur_cmd_buf) |_, i| {
            var buf = &ctx.cur_cmd_buf[i];
            assert(!buf.in_render_pass);
            if (buf.queue_transfer_src) {
                assert(!buf.queue_transfer_dst);
                _ = try buf.submit(device, null, .{}, null);
                assert(!buf.queue_transfer_src);
                submits[i] = true;
            }
        }

        for (ctx.cur_cmd_buf) |*cmd, i| {
            cmd.queue_transfer_dst = false;
            if (submits[i]) {
                _ = try get(@intToEnum(queues.QueueId, i), device); // reinitialize cmdbuf
            }
        }
    }
};

pub const Pool = struct {
    const Self = @This();

    handle: vk.CommandPool,

    pub const InitError = DeviceDispatch.CreateCommandPoolError;
    pub fn init(device: *const Device, family: u32, flags: vk.CommandPoolCreateFlags) InitError!Self {
        const handle = try device.dispatch.createCommandPool(device.handle, &.{
            .flags = flags,
            .queue_family_index = family,
        }, null);

        return Self{
            .handle = handle,
        };
    }

    pub fn deinit(self: *const Self, device: *const Device) void {
        device.dispatch.destroyCommandPool(device.handle, self.handle, null);
    }

    pub fn reset(self: *const Self, device: *const Device, flags: vk.CommandPoolResetFlagBits) !void {
        try device.dispatch.resetCommandPool(device.handle, self.handle, flags);
    }
};
