const assert = @import("std").debug.assert;
const vk = @import("vulkan");

const Device = @import("device.zig").Device;
const Swapchain = @import("swapchain.zig").Swapchain;
const settings = @import("settings.zig");
const queues = @import("queues.zig");
const Image = @import("image.zig").Image;

const Self = @This();

width: u32,
height: u32,
depth: [settings.resource_sets]Image,
scene: [settings.resource_sets]Image,

// TODO: scaled render resolution
pub fn getDesiredWidth(swapchain: *Swapchain) u32 {
    return getDisplayWidth(swapchain);
}

pub fn getDesiredHeight(swapchain: *Swapchain) u32 {
    return getDisplayHeight(swapchain);
}

var s_targets: Self = undefined;

pub fn init(device: *const Device, swapchain: *Swapchain) !void {
    errdefer deinit(device);

    const width = getDesiredWidth(swapchain);
    const height = getDesiredHeight(swapchain);

    s_targets.width = width;
    s_targets.height = height;

    for (s_targets.scene) |*s| {
        const usage: vk.ImageUsageFlags = .{
            .color_attachment_bit = true,
            .sampled_bit = true,
            .storage_bit = true,
            .transfer_src_bit = true,
            .transfer_dst_bit = true,
        };
        const queue_family_indices = [_]u32{
            queues.get(.graphics).family,
            queues.get(.compute).family,
        };

        try Image.init(s, device, .{
            .flags = .{},
            .image_type = .@"2d",
            .format = .r16g16b16a16_sfloat,
            .extent = .{
                .width = width,
                .height = height,
                .depth = 1,
            },
            .mip_levels = 1,
            .array_layers = 1,
            .samples = .{ .@"1_bit" = true },
            .tiling = .optimal,
            .usage = usage,
            .sharing_mode = .exclusive,
            .queue_family_index_count = queue_family_indices.len,
            .p_queue_family_indices = &queue_family_indices,
            .initial_layout = .@"undefined",
        }, .gpuOnly);
    }

    for (s_targets.depth) |*d| {
        const usage: vk.ImageUsageFlags = .{
            .depth_stencil_attachment_bit = true,
            .sampled_bit = true,
        };
        const queue_family_indices = [_]u32{
            queues.get(.graphics).family,
        };
        try Image.init(d, device, .{
            .flags = .{},
            .image_type = .@"2d",
            .format = .x8_d24_unorm_pack32,
            .extent = .{
                .width = width,
                .height = height,
                .depth = 1,
            },
            .mip_levels = 1,
            .array_layers = 1,
            .samples = .{ .@"1_bit" = true },
            .tiling = .optimal,
            .usage = usage,
            .sharing_mode = .exclusive,
            .queue_family_index_count = queue_family_indices.len,
            .p_queue_family_indices = &queue_family_indices,
            .initial_layout = .@"undefined",
        }, .gpuOnly);
    }
}

pub fn deinit(device: *const Device) void {
    for (s_targets.scene) |*s| {
        s.deinit(device);
    }

    for (s_targets.depth) |*d| {
        d.deinit(device);
    }
}

pub fn recreate(device: *const Device) void {
    if (getRenderWidth() != getDesiredWidth() or
        getRenderHeight() != getDesiredHeight())
    {
        deinit(device);
        init();
    }
}

pub fn getDisplayWidth(swapchain: *const Swapchain) u32 {
    return swapchain.getBackBufferRO().width;
}

pub fn getDisplayHeight(swapchain: *const Swapchain) u32 {
    return swapchain.getBackBufferRO().height;
}

pub fn getDepthBuffer(swapchain: *Swapchain) *Image {
    const img = &s_targets.depth[swapchain.sync_index];
    assert(img.handle != .null_handle);
    return img;
}

pub fn getSceneBuffer(swapchain: *Swapchain) *Image {
    var img = &s_targets.scene[swapchain.sync_index];
    assert(img.handle != .null_handle);
    return img;
}

pub fn getRenderWidth() u32 {
    return s_targets.width;
}

pub fn getRenderHeight() u32 {
    return s_targets.height;
}

pub fn getRenderAspect() f32 {
    return @intToFloat(f32, s_targets.width) / @intToFloat(f32, s_targets.height);
}

pub fn getRenderScale() f32 {
    return 1.0; // TODO convar
}
