const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");
const Image = @import("image.zig").Image;
const RenderPass = @import("RenderPass.zig");
const Device = @import("device.zig").Device;

// TODO memory
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const allocator = gpa.allocator();

const num_attachments = 8;

const Key = struct {
    attachments: [num_attachments]vk.ImageView = std.mem.zeroes([num_attachments]vk.ImageView),
    formats: [num_attachments]vk.Format = std.mem.zeroes([num_attachments]vk.Format),
    width: u16,
    height: u16,
};

const Value = struct {
    handle: vk.Framebuffer,
};

const Self = @This();

const FramebufHashMap = std.AutoArrayHashMap(Key, Value);
var s_fbufs: FramebufHashMap = undefined;

pub fn init() !void {
    s_fbufs = FramebufHashMap.init(allocator);
    try s_fbufs.ensureTotalCapacity(16);
}

pub fn deinit(device: *const Device) void {
    for (s_fbufs.values()) |*entry| {
        device.dispatch.destroyFramebuffer(device.handle, entry.handle, null);
    }

    s_fbufs.deinit();
}

pub fn getOrAdd(device: *const Device, attachments: []*const Image, width: u32, height: u32) !vk.Framebuffer {
    std.debug.assert(attachments.len >= 1);
    std.debug.assert(attachments.len <= num_attachments);

    var key = Key{
        .width = @intCast(u16, width),
        .height = @intCast(u16, height),
    };

    for (attachments) |attachment, i| {
        assert(attachment.view != .null_handle);
        key.attachments[i] = attachment.view;
        key.formats[i] = attachment.format;
    }

    const result = s_fbufs.getOrPutAssumeCapacity(key);
    if (!result.found_existing) {
        std.debug.print("new fbuf: att {any} \n\n format {any}\n\n", .{ key.attachments[0..attachments.len], key.formats[0..attachments.len] });
        const value = Value{
            .handle = try Framebuffer.init(device, key.attachments[0..attachments.len], key.formats[0..attachments.len], key.width, key.height),
        };
        result.value_ptr.* = value;
    }

    return result.value_ptr.handle;
}

pub fn remove(device: *const Device, view: vk.ImageView) void {
    var iter = s_fbufs.iterator();
    while (iter.next()) |entry| {
        if (containsView(entry.key_ptr, view)) {
            Framebuffer.deinit(entry.value_ptr.handle, device);
            _ = s_fbufs.swapRemove(entry.key_ptr.*);
        }
    }
}

fn containsView(key: *const Key, view: vk.ImageView) bool {
    for (key.attachments) |attachment| {
        if (attachment == view) {
            return true;
        }
    }

    return false;
}

pub const Framebuffer = struct {
    pub fn init(device: *const Device, attachments: []const vk.ImageView, formats: []const vk.Format, width: u32, height: u32) !vk.Framebuffer {
        var pass_desc: RenderPass.Description = .{
            .src_access_mask = .{ .shader_read_bit = true },
            .dst_access_mask = .{ .color_attachment_write_bit = true },
            .src_stage_mask = .{ .fragment_shader_bit = true },
            .dst_stage_mask = .{ .color_attachment_output_bit = true },
        };

        for (formats) |format, i| {
            pass_desc.attachments[i].format = format;
            const layout = if (i == 0 and @enumToInt(format) >= @enumToInt(vk.Format.d16_unorm) and @enumToInt(format) <= @enumToInt(vk.Format.d32_sfloat_s8_uint))
                vk.ImageLayout.depth_stencil_attachment_optimal
            else
                vk.ImageLayout.color_attachment_optimal;
            pass_desc.attachments[i].layout = layout;
        }

        const pass = try RenderPass.get(device, pass_desc);
        return try device.dispatch.createFramebuffer(device.handle, &.{
            .flags = .{},
            .render_pass = pass,
            .attachment_count = @intCast(u32, attachments.len),
            .p_attachments = attachments.ptr,
            .width = width,
            .height = height,
            .layers = 1,
        }, null);
    }

    pub fn deinit(framebuffer: vk.Framebuffer, device: *const Device) void {
        device.dispatch.destroyFramebuffer(device.handle, framebuffer, null);
    }
};
