const builtin = @import("builtin");
const std = @import("std");
const assert = std.debug.assert;

const vk = @import("vulkan");
const vma = @import("vma");
const Device = @import("device.zig").Device;
const Command = @import("Command.zig");
const SubmitId = @import("submit_id.zig").SubmitId;
const queues = @import("queues.zig");
const settings = @import("settings.zig");
const memory = @import("memory.zig");
const Swapchain = @import("swapchain.zig").Swapchain;

// TODO memory
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const allocator = gpa.allocator();

pub const Image = struct {
    const Self = @This();

    state: State = .{},
    handle: vk.Image = .null_handle,
    allocation: vma.Allocation,
    view: vk.ImageView = .null_handle,
    format: vk.Format,
    width: u16,
    height: u16,
    depth: u12,
    mip_levels: u8,
    array_layers: u8,
    usage: vk.ImageUsageFlags,
    image_type: vk.ImageType,
    imported: bool,

    pub fn init(self: *Self, device: *const Device, info: vk.ImageCreateInfo, mem_usage: vma.MemoryUsage) !void {
        try memory.imageNew(self, device, info, mem_usage);
    }

    pub fn release(self: *const Self) !void {
        self.getSubmit();
    }

    pub fn deinit(self: *const Self, device: *const Device) void {
        memory.imageDel(self, device);
    }

    /// import vk.Image handle into an existing Image object.
    pub fn import(self: *Image, device: *const Device, info: vk.ImageCreateInfo, handle: vk.Image) !void {
        errdefer memory.imageDel(self, device);

        self.handle = handle;
        self.image_type = info.image_type;
        self.format = info.format;
        self.state = State{
            .layout = info.initial_layout,
            .stage = .{ .color_attachment_output_bit = true }, // possibly needed
        };
        self.usage = info.usage;
        self.width = @intCast(u16, info.extent.width);
        self.height = @intCast(u16, info.extent.height);
        self.depth = @intCast(u12, info.extent.depth);
        self.mip_levels = @intCast(u8, info.mip_levels);
        self.array_layers = @intCast(u8, info.array_layers);
        self.imported = true;

        var view_info = try infoToViewInfo(&info);
        view_info.image = self.handle;
        self.view = try device.dispatch.createImageView(device.handle, &view_info, null);
    }

    fn getSubmit(self: *Self) SubmitId {
        assert(!self.state.substates);
        if (self.handle != .null_handle) {
            const id = if (self.state.stage != 0) {
                .{
                    .counter = self.state.cmd_id,
                    .queue_id = self.state.owner,
                    .valid = true,
                };
            } else {
                Command.getHeadSubmit(self.state.owner);
            };
            assert(id.valid);
            return id;
        }

        return SubmitId{};
    }

    pub const State = struct {
        owner: queues.QueueId = .graphics,
        cmd_id: u32 = 0,
        stage: vk.PipelineStageFlags = .{},
        access: vk.AccessFlags = .{},
        layout: vk.ImageLayout = .@"undefined",
        substates: ?[]SubState = null,

        pub fn transferSrc(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.xfer);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .transfer_bit = true },
                .access = .{ .transfer_read_bit = true },
                .layout = .transfer_src_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn transferDst(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.xfer);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .transfer_bit = true },
                .access = .{ .transfer_write_bit = true },
                .layout = .transfer_dst_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn fragSample(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn computeSample(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.comp);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn fragLoad(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn fragStore(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_write_bit = true },
                .layout = .general,
            };
            return try state.handle(device, buf, img);
        }

        pub fn fragLoadStore(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true, .shader_write_bit = true },
                .layout = .general,
            };
            return try state.handle(device, buf, img);
        }

        pub fn computeLoad(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.comp);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn computeStore(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.comp);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_write_bit = true },
                .layout = .general,
            };
            return try state.handle(device, buf, img);
        }

        pub fn computeLoadStore(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.comp);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true, .shader_write_bit = true },
                .layout = .general,
            };
            return try state.handle(device, buf, img);
        }

        pub fn colorAttachWrite(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .color_attachment_output_bit = true },
                .access = .{ .color_attachment_write_bit = true },
                .layout = .color_attachment_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn depthAttachWrite(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.gfx);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .late_fragment_tests_bit = true },
                .access = .{ .depth_stencil_attachment_write_bit = true },
                .layout = .depth_stencil_attachment_optimal,
            };
            return try state.handle(device, buf, img);
        }

        pub fn presentSrc(device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(buf.pres);
            const state = State{
                .owner = buf.queue_id,
                .stage = .{ .all_graphics_bit = true },
                .access = .{ .color_attachment_read_bit = true, .color_attachment_write_bit = true },
                .layout = .present_src_khr,
            };
            return try state.handle(device, buf, img);
        }

        fn handle(self: *const State, device: *const Device, buf: *Command.Buffer, img: *Image) !bool {
            assert(img.handle != .null_handle);
            assert(self.access.toInt() != 0);
            assert(self.stage.toInt() != 0);

            var inserted_barrier = false;
            const prev = &img.state;
            const prev_queue = queues.get(prev.owner);
            const next_queue = queues.get(self.owner);

            assert(self.stage.intersect(next_queue.stage_mask).toInt() == self.stage.toInt());
            assert(self.access.intersect(next_queue.access_mask).toInt() == self.access.toInt());

            if (prev.substates != null) {
                assert(prev.owner == self.owner);
                inserted_barrier = try SubState.handle(device, buf, img, &.{
                    .stage = self.stage,
                    .access = self.access,
                    .layout = self.layout,
                }, 0, img.array_layers, 0, img.mip_levels);
                assert(img.state.substates == null);
                return inserted_barrier;
            }

            var new_resource = false;
            if (prev.stage.toInt() == 0) {
                prev.stage = .{ .top_of_pipe_bit = true };
                new_resource = true;
            }

            const aspect = if (img.usage.depth_stencil_attachment_bit) vk.ImageAspectFlags{ .depth_bit = true } else vk.ImageAspectFlags{ .color_bit = true };

            if (prev_queue != next_queue) {
                // queue ownership transfer
                var src_cmd = try Command.Buffer.get(prev.owner, device);
                var dst_cmd = try Command.Buffer.get(self.owner, device);
                if (src_cmd.queue_transfer_dst or dst_cmd.queue_transfer_src) {
                    try Command.Buffer.flushQueueTransfers(device);
                }

                const barrier = vk.ImageMemoryBarrier{
                    .src_access_mask = prev.access,
                    .dst_access_mask = self.access,
                    .old_layout = prev.layout,
                    .new_layout = self.layout,
                    .src_queue_family_index = prev_queue.family,
                    .dst_queue_family_index = next_queue.family,
                    .image = img.handle,
                    .subresource_range = .{
                        .aspect_mask = aspect,
                        .base_mip_level = 0,
                        .level_count = vk.REMAINING_MIP_LEVELS,
                        .base_array_layer = 0,
                        .layer_count = vk.REMAINING_ARRAY_LAYERS,
                    },
                };

                assert(!src_cmd.queue_transfer_dst);
                src_cmd.queue_transfer_src = true;
                device.dispatch.cmdPipelineBarrier(src_cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 0, undefined, 1, @ptrCast([*]const vk.ImageMemoryBarrier, &barrier));
                assert(!dst_cmd.queue_transfer_src);
                dst_cmd.queue_transfer_dst = true;
                device.dispatch.cmdPipelineBarrier(dst_cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 0, undefined, 1, @ptrCast([*]const vk.ImageMemoryBarrier, &barrier));
                prev.* = self.*;
                inserted_barrier = true;
            } else {
                const layout_change = prev.layout != self.layout;
                const src_read = prev.access.intersect(Command.read_access).toInt() != 0;
                const src_write = prev.access.intersect(Command.write_access).toInt() != 0;
                const dst_read = self.access.intersect(Command.read_access).toInt() != 0;
                const dst_write = self.access.intersect(Command.write_access).toInt() != 0;
                const read_after_write = src_write and dst_read;
                const write_after_read = src_read and dst_write;
                const write_after_write = src_write and dst_write;
                if (layout_change or read_after_write or write_after_read or write_after_write or new_resource) {
                    // data hazard, insert barrier
                    const cmd = try Command.Buffer.get(self.owner, device);
                    const barrier = vk.ImageMemoryBarrier{
                        .src_access_mask = prev.access,
                        .dst_access_mask = self.access,
                        .old_layout = prev.layout,
                        .new_layout = self.layout,
                        .src_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                        .dst_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                        .image = img.handle,
                        .subresource_range = .{
                            .aspect_mask = aspect,
                            .base_mip_level = 0,
                            .level_count = vk.REMAINING_MIP_LEVELS,
                            .base_array_layer = 0,
                            .layer_count = vk.REMAINING_ARRAY_LAYERS,
                        },
                    };
                    device.dispatch.cmdPipelineBarrier(cmd.handle, prev.stage, self.stage, .{}, 0, undefined, 0, undefined, 1, @ptrCast([*]const vk.ImageMemoryBarrier, &barrier));
                    prev.* = self.*;
                    inserted_barrier = true;
                } else {
                    // no data hazard, append usage state
                    prev.stage = prev.stage.merge(self.stage);
                    prev.access = prev.access.merge(self.access);
                }
            }

            (try Command.Buffer.get(self.owner, device)).touchImage(img);
            assert(img.state.layout == self.layout);

            return inserted_barrier;
        }
    };

    const SubState = struct {
        stage: vk.PipelineStageFlags,
        access: vk.AccessFlags,
        layout: vk.ImageLayout,

        pub fn transferSrc(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.xfer);
            const self = SubState{
                .stage = .{ .transfer_bit = true },
                .access = .{ .transfer_read_bit = true },
                .layout = .transfer_src_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn transferDst(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.xfer);
            const self = SubState{
                .stage = .{ .transfer_bit = true },
                .access = .{ .transfer_write_bit = true },
                .layout = .transfer_dst_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn fragSample(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.gfx);
            const self = SubState{
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn computeSample(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.comp);
            const self = SubState{
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .shader_read_only_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn fragLoad(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.gfx);
            const self = SubState{
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn fragStore(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.xfer);
            const self = SubState{
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_write_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn fragLoadStore(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.gfx);
            const self = SubState{
                .stage = .{ .fragment_shader_bit = true },
                .access = .{ .shader_read_bit = true, .shader_write_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn computeLoad(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.compm);
            const self = SubState{
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn computeStore(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.comp);
            const self = SubState{
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_write_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn computeLoadStore(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.comp);
            const self = SubState{
                .stage = .{ .compute_shader_bit = true },
                .access = .{ .shader_read_bit = true, .shader_write_bit = true },
                .layout = .general,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn colorAttachWrite(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.gfx);
            const self = SubState{
                .stage = .{ .color_attachment_output_bit = true },
                .access = .{ .color_attachment_write_bit = true },
                .layout = .color_attachment_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn depthAttachWrite(device: *const Device, buf: *Command.Buffer, img: *Image, base_layer: i32, layer_count: i32, base_mip: i32, mip_count: i32) !bool {
            assert(buf.gfx);
            const self = SubState{
                .stage = .{ .late_fragment_tests_bit = true },
                .access = .{ .depth_stencil_attachment_write_bit = true },
                .layout = .depth_stencil_attachment_optimal,
            };
            return try handle(device, buf, img, &self, base_layer, layer_count, base_mip, mip_count);
        }

        pub fn handle(device: *const Device, buf: *Command.Buffer, img: *Image, next: *const SubState, base_layer: usize, layer_count: usize, base_mip: usize, mip_count: usize) !bool {
            {
                assert(img.handle != .null_handle);
                assert(next.access.toInt() != 0);
                assert(next.stage.toInt() != 0);
                assert(layer_count > 0);
                assert(mip_count > 0);
                assert(buf == (try Command.Buffer.get(img.state.owner, device)));
                assert(buf.handle != .null_handle);
                if (builtin.mode == .Debug or builtin.mode == .ReleaseSafe) {
                    const queue = queues.get(buf.queue_id);
                    assert(next.stage.toInt() == queue.stage_mask.toInt());
                    assert(next.access.toInt() == queue.access_mask.toInt());
                }
            }

            var inserted_barrier = false;
            const L = img.array_layers;
            const M = img.mip_levels;
            assert(base_layer + layer_count <= L);
            assert(base_mip + mip_count <= M);
            const substate_count = L * M;

            var prev = &img.state;
            var substates: []SubState = undefined;
            if (prev.substates) |ss| {
                assert(prev.layout == .@"undefined");
                assert(prev.access.toInt() == 0);
                assert(prev.stage.toInt() == 0);
                substates = ss;
            } else {
                substates = try allocator.alloc(SubState, substate_count);
                prev.substates = substates;
                if (prev.stage.toInt() == 0) {
                    prev.stage = .{ .top_of_pipe_bit = true };
                }

                for (substates) |*ss| {
                    ss.stage = prev.stage;
                    ss.access = prev.access;
                    ss.layout = prev.layout;
                }
            }

            const aspect = if (img.usage.depth_stencil_attachment_bit) vk.ImageAspectFlags{ .depth_bit = true } else vk.ImageAspectFlags{ .color_bit = true };
            const max_barriers = layer_count * mip_count;
            const barriers = try allocator.alloc(vk.ImageMemoryBarrier, max_barriers);
            defer allocator.free(barriers);

            prev.layout = .@"undefined";
            prev.access = .{};
            prev.stage = .{};

            var b: usize = 0;
            var l: usize = 0;
            while (l < layer_count) : (l += 1) {
                const i_layer = base_layer + l;
                var m: usize = 0;
                while (m < mip_count) : (m += 1) {
                    const i_mip = base_mip + m;
                    const prev_sub = &substates[i_layer * M + i_mip];
                    prev.stage = prev.stage.merge(prev_sub.stage);
                    prev.access = prev.access.merge(prev_sub.access);

                    const layout_change = prev_sub.layout != next.layout;
                    const src_read = prev_sub.access.contains(Command.read_access);
                    const src_write = prev_sub.access.contains(Command.write_access);
                    const dst_read = next.access.contains(Command.read_access);
                    const dst_write = next.access.contains(Command.write_access);
                    const read_after_write = src_write and dst_read;
                    const write_after_read = src_read and dst_write;
                    const write_after_write = src_write and dst_write;

                    if (layout_change or read_after_write or write_after_read or write_after_write) {
                        assert(b < max_barriers);
                        const barrier = &barriers[b];
                        b += 1;
                        barrier.* = .{
                            .src_access_mask = prev_sub.access,
                            .dst_access_mask = next.access,
                            .old_layout = prev_sub.layout,
                            .new_layout = next.layout,
                            .src_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                            .dst_queue_family_index = vk.QUEUE_FAMILY_IGNORED,
                            .image = img.handle,
                            .subresource_range = .{
                                .aspect_mask = aspect,
                                .base_mip_level = @intCast(u32, i_mip),
                                .level_count = 1,
                                .base_array_layer = @intCast(u32, i_layer),
                                .layer_count = 1,
                            },
                        };
                        prev_sub.* = next.*;
                    } else {
                        prev_sub.access = prev_sub.access.merge(next.access);
                        prev_sub.stage = prev_sub.stage.merge(next.stage);
                    }
                }
            }

            // merge mips
            var i: usize = 0;
            while ((i + 1) < b) : (i += 1) {
                const dst = &barriers[i];
                const src = &barriers[i + 1];
                const dst_range = &dst.subresource_range;
                const src_range = &src.subresource_range;
                if (dst.old_layout != src.old_layout) {
                    continue;
                }

                if (dst_range.base_array_layer != src_range.base_array_layer) {
                    continue;
                }

                if (dst_range.base_mip_level + dst_range.level_count == src_range.base_mip_level) {
                    dst_range.level_count += src_range.level_count;
                    dst.src_access_mask = dst.src_access_mask.merge(src.src_access_mask);
                    assert(dst_range.level_count <= img.mip_levels);
                    var j: usize = i + 1;
                    while ((j + 1) < b) : (j += 1) {
                        barriers[j] = barriers[j + 1];
                    }

                    b -= 1;
                    assert(b >= 1);
                    i -= 1;
                }
            }

            // merge layers
            i = 0;
            while ((i + 1) < b) : (i += 1) {
                const dst = &barriers[i];
                const src = &barriers[i + 1];
                const dst_range = &dst.subresource_range;
                const src_range = &src.subresource_range;
                if (dst.old_layout != src.old_layout) {
                    continue;
                }

                if (dst_range.base_mip_level != src_range.base_mip_level) {
                    continue;
                }

                if (dst_range.level_count != src_range.level_count) {
                    continue;
                }

                if (dst_range.base_array_layer + dst_range.layer_count == src_range.base_array_layer) {
                    dst_range.layer_count += src_range.layer_count;
                    dst.src_access_mask = dst.src_access_mask.merge(src.src_access_mask);
                    assert(dst_range.layer_count <= img.array_layers);
                    var j: usize = i + 1;
                    while ((j + 1) < b) : (j += 1) {
                        barriers[j] = barriers[j + 1];
                    }

                    b -= 1;
                    assert(b >= 1);
                    i -= 1;
                }
            }

            assert(b >= 1);
            device.dispatch.cmdPipelineBarrier(buf.handle, prev.stage, next.stage, .{}, 0, undefined, 0, undefined, @intCast(u32, b), barriers.ptr);
            inserted_barrier = true;

            b = 0;

            prev.access = substates[0].access;
            prev.stage = substates[0].stage;
            prev.layout = substates[0].layout;
            var is_uniform = true;
            i = 1;
            while (i < substate_count) : (i += 1) {
                if (substates[0].layout != substates[i].layout) {
                    is_uniform = false;
                    break;
                }

                prev.access = prev.access.merge(substates[i].access);
                prev.stage = prev.stage.merge(substates[i].stage);
            }

            if (is_uniform) {
                prev.substates = null;
                allocator.free(substates);
            } else {
                prev.layout = .@"undefined";
                prev.access = .{};
                prev.stage = .{};
            }

            buf.touchImage(img);

            return inserted_barrier;
        }
    };

    const Set = struct {
        frames: [settings.resource_sets]Image,

        pub fn init(info: *const vk.ImageCreateInfo, mem_usage: vma.MemUsage) !Set {
            const self = Set{};

            errdefer self.deinit();
            for (self.frames) |frame| {
                try frame.init(info, mem_usage);
            }
        }

        pub fn deinit(self: *Set) void {
            for (self.frames) |frame| {
                frame.deinit();
            }
        }

        pub fn reserve(self: *Set, swapchain: *const Swapchain, info: *const vk.ImageCreateInfo, mem_usage: memory.Usage) !void {
            try self.current(swapchain).reserve(info, mem_usage);
        }

        pub fn current(self: *Set, swapchain: *const Swapchain) *Image {
            const sync_index = swapchain.getSyncIndex();
            assert(sync_index < self.frames.len);
            return &self.frames[sync_index];
        }

        pub fn next(self: *Set, swapchain: *const Swapchain) *Image {
            const next_index = swapchain.getNextSyncIndex();
            assert(next_index < self.frames.len);
            return &self.frames[next_index];
        }

        pub fn prev(self: *Set, swapchain: *const Swapchain) *Image {
            const prev_index = swapchain.getPrevSyncIndex();
            assert(prev_index < self.frames.len);
            return &self.frames[prev_index];
        }
    };
};

pub fn infoToViewInfo(info: *const vk.ImageCreateInfo) !vk.ImageViewCreateInfo {
    // anything but transfer usage needs a view
    const viewless: vk.ImageUsageFlags = .{ .transfer_src_bit = true, .transfer_dst_bit = true };
    if (info.usage.intersect(viewless.complement()).toInt() != 0) {
        return vk.ImageViewCreateInfo{
            .flags = .{},
            .format = info.format,
            .view_type = infoToViewType(info),
            .image = .null_handle,
            .components = .{ .r = .identity, .g = .identity, .b = .identity, .a = .identity },
            .subresource_range = .{
                .aspect_mask = infoToAspects(info),
                .base_mip_level = 0,
                .level_count = info.mip_levels,
                .base_array_layer = 0,
                .layer_count = info.array_layers,
            },
        };
    }

    // Unsure if this should be error or not
    return error.NotConvertibleToViewInfo;
}

pub fn infoToViewType(info: *const vk.ImageCreateInfo) vk.ImageViewType {
    return switch (info.image_type) {
        .@"1d" => {
            return if (info.array_layers <= 1) .@"1d" else .@"1d_array";
        },
        .@"3d" => {
            return .@"3d";
        },
        else => {
            if (info.array_layers <= 1) {
                return .@"2d";
            } else if (info.array_layers == 6) {
                return .cube;
            } else {
                return .@"2d_array";
            }
        },
    };
}

pub fn infoToAspects(info: *const vk.ImageCreateInfo) vk.ImageAspectFlags {
    return switch (info.format) {
        .d16_unorm, .x8_d24_unorm_pack32, .d32_sfloat => .{ .depth_bit = true },
        .d16_unorm_s8_uint, .d24_unorm_s8_uint, .d32_sfloat_s8_uint => .{ .depth_bit = true, .stencil_bit = true },
        .s8_uint => .{ .stencil_bit = true },
        else => .{ .color_bit = true },
    };
}
