const std = @import("std");
const vk = @import("vulkan");
const assert = std.debug.assert;

const ProfileMark = @import("../../common/profiler.zig").ProfileMark;
const time = @import("../../common/time.zig");

const framebuffer = @import("framebuffer.zig");
const Base = @import("Base.zig");
const Device = @import("device.zig").Device;
const Command = @import("Command.zig");
const Instance = @import("instance.zig").Instance;
const SubmitId = @import("submit_id.zig").SubmitId;
const settings = @import("settings.zig");
const queues = @import("queues.zig");
const vma = @import("vma");
const image = @import("image.zig");
const Buffer = @import("Buffer.zig");

// TODO memory
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const perm_allocator = gpa.allocator();

var s_allocator: Allocator = undefined;

pub fn init(instance: *const Instance, device: *const Device) !void {
    s_allocator = try Allocator.init(instance, device);
}

pub fn deinit(device: *const Device) void {
    finalize(device) catch {};
    s_allocator.deinit();
}

pub fn finalize(device: *const Device) !void {
    try device.waitIdle();

    s_allocator.mutex.lock();
    defer s_allocator.mutex.unlock();

    for (s_allocator.releasables.items) |releasable| {
        try releasable.submit_id.wait(device);
        releasable.deinit(device, s_allocator);
    }
    s_allocator.releasables.clearRetainingCapacity();
}

const pm_map = ProfileMark.init("memory.map");
pub fn map(allocation: vma.Allocation) ![*]u8 {
    pm_map.begin();
    defer pm_map.end();

    assert(allocation != .Null);
    return try s_allocator.handle.mapMemory(allocation, u8);
}

const pm_unmap = ProfileMark.init("memory.unmap");
pub fn unmap(allocation: vma.Allocation) void {
    pm_unmap.begin();
    defer pm_unmap.end();

    assert(allocation != .Null);
    s_allocator.handle.unmapMemory(allocation);
}

const pm_flush = ProfileMark.init("memory.flush");
pub fn flush(allocation: vma.Allocation) void {
    pm_flush.begin();
    defer pm_flush.end();

    s_allocator.handle.flushAllocation(allocation, 0, vk.WHOLE_SIZE);
    s_allocator.handle.invalidateAllocation(allocation, 0, vk.WHOLE_SIZE);
}

const UpdateError = SubmitId.PollError || FinalizeCheckError;
const pm_update = ProfileMark.init("memory.update");
pub fn update(device: *const Device) UpdateError!void {
    pm_update.begin();
    defer pm_update.end();

    if (s_allocator.handle != .Null) {
        const frame = time.getFrameCount();
        s_allocator.handle.setCurrentFrameIndex(@intCast(u32, frame)); // XXX hmm
        {
            s_allocator.mutex.lock();
            defer s_allocator.mutex.unlock();

            if (s_allocator.releasables.items.len >= 1) {
                var i = s_allocator.releasables.items.len - 1;
                while (i >= 0) : (i -= 1) {
                    if (try s_allocator.releasables.items[i].submit_id.poll(device)) {
                        _ = s_allocator.releasables.swapRemove(i);
                    }
                }
            }
        }

        try finalizeCheck(device, s_allocator.releasables.items.len);
    }
}

const pm_imgnew = ProfileMark.init("memory.imageNew");
pub fn imageNew(img: *image.Image, device: *const Device, info: vk.ImageCreateInfo, mem_usage: vma.MemoryUsage) !void {
    pm_imgnew.begin();
    defer pm_imgnew.end();

    assert(s_allocator.handle != .Null);

    errdefer |err| {
        std.debug.print("Failed to allocate image: {}\n", .{err});
        std.debug.print("Size: {} x {} x {}\n", .{ info.extent.width, info.extent.height, info.extent.depth });
        std.debug.print("Mip Levels: {}\n", .{info.mip_levels});
        std.debug.print("Array Layers: {}\n", .{info.array_layers});
        imageDel(img, device);
    }

    // memset needed?
    img.image_type = info.image_type;
    img.format = info.format;
    img.state.owner = .graphics;
    img.state.stage = .{};
    img.state.layout = info.initial_layout;
    img.usage = info.usage;
    img.width = @intCast(u16, info.extent.width);
    img.height = @intCast(u16, info.extent.height);
    img.depth = @intCast(u12, info.extent.depth);
    img.mip_levels = @intCast(u8, info.mip_levels);
    img.array_layers = @intCast(u8, info.array_layers);
    img.imported = false;

    const pool = getTexturePool(info.usage, mem_usage) orelse return error.NoPoolForImage;

    const result = try s_allocator.handle.createImage(info, .{
        .flags = .{
            .withinBudget = true,
        },
        .usage = mem_usage,
        .pool = pool,
    });
    img.handle = result.image;
    img.allocation = result.allocation;

    var view_info = try image.infoToViewInfo(&info);
    view_info.image = img.handle;
    img.view = try device.dispatch.createImageView(device.handle, &view_info, null);
}

const pm_imagedel = ProfileMark.init("memory.imageDel");
pub fn imageDel(img: *const image.Image, device: *const Device) void {
    pm_imagedel.begin();
    defer pm_imagedel.end();

    assert(s_allocator.handle != .Null);

    device.dispatch.destroyImageView(device.handle, img.view, null);
    vma.Allocator.destroyImage(s_allocator.handle, img.handle, img.allocation);
}

const pm_bufnew = ProfileMark.init("memory.bufferNew");
pub fn bufferNew(buf: *Buffer, size: usize, usage: vk.BufferUsageFlags, mem_usage: vma.MemoryUsage) !void {
    pm_bufnew.begin();
    defer pm_bufnew.end();

    assert(s_allocator.handle != .Null);
    assert(size > 0);

    buf.state = std.mem.zeroes(Buffer.State);
    buf.state.owner = .graphics;

    const buffer_info = vk.BufferCreateInfo{
        .flags = .{},
        .size = size,
        .usage = usage,
        .sharing_mode = .exclusive,
        .queue_family_index_count = 0,
        .p_queue_family_indices = undefined,
    };
    const alloc_info = vma.AllocationCreateInfo{
        .flags = .{ .withinBudget = true },
        .usage = mem_usage,
        .pool = try getBufferPool(usage, mem_usage),
    };

    const result = try s_allocator.handle.createBuffer(buffer_info, alloc_info);
    buf.handle = result.buffer;
    buf.allocation = result.allocation;
    buf.size = size;
}

fn getTexturePool(usage: vk.ImageUsageFlags, mem_usage: vma.MemoryUsage) ?vma.Pool {
    var pool: ?vma.Pool = null;
    switch (mem_usage) {
        .gpuOnly => {
            const pool_usage = s_allocator.device_texture_pool.image_usage orelse unreachable;
            if (usage.intersect(pool_usage).toInt() != 0) {
                pool = s_allocator.device_texture_pool.handle;
            }
        },
        else => {},
    }

    const attachment_usage: vk.ImageUsageFlags = .{ .color_attachment_bit = true, .depth_stencil_attachment_bit = true };

    std.debug.assert(pool != null or usage.intersect(attachment_usage).toInt() != 0);
    return pool;
}

fn getBufferPool(usage: vk.BufferUsageFlags, mem_usage: vma.MemoryUsage) !vma.Pool {
    switch (mem_usage) {
        .cpuOnly => {
            const pool_usage = s_allocator.staging_pool.buffer_usage orelse unreachable;
            assert(usage.intersect(pool_usage).toInt() != 0);
            return s_allocator.staging_pool.handle;
        },
        .gpuToCpu, .cpuToGpu => { // just a guess
            const pool_usage = s_allocator.dynamic_buffer_pool.buffer_usage orelse unreachable;
            assert(usage.intersect(pool_usage).toInt() != 0);
            return s_allocator.dynamic_buffer_pool.handle;
        },
        .gpuOnly => {
            const pool_usage = s_allocator.device_buffer_pool.buffer_usage orelse unreachable;
            assert(usage.intersect(pool_usage).toInt() != 0);
            return s_allocator.device_buffer_pool.handle;
        },
        else => return error.InvalidMemoryUsage,
    }
}

const Pool = struct {
    const Self = @This();

    handle: vma.Pool,
    size: vk.DeviceSize,
    buffer_usage: ?vk.BufferUsageFlags = null,
    image_usage: ?vk.ImageUsageFlags = null,
    mem_usage: vma.MemoryUsage,
    queue_usage: queues.QueueFlags,

    pub fn init(handle: vma.Allocator, size: vk.DeviceSize, buffer_usage: ?vk.BufferUsageFlags, image_usage: ?vk.ImageUsageFlags, mem_usage: vma.MemoryUsage, queue_usage: queues.QueueFlags) !Self {
        std.debug.assert(size > 0);
        std.debug.assert((buffer_usage != null and image_usage == null) or (buffer_usage == null and image_usage != null));

        var queue_family_count: usize = 0;
        var queue_families: [queues.QueueId.count]u32 = undefined;
        if (queue_usage.graphics_bit) {
            queue_families[queue_family_count] = queues.get(queues.QueueId.graphics).family;
            queue_family_count += 1;
        }

        if (queue_usage.compute_bit) {
            queue_families[queue_family_count] = queues.get(queues.QueueId.compute).family;
            queue_family_count += 1;
        }

        if (queue_usage.transfer_bit) {
            queue_families[queue_family_count] = queues.get(queues.QueueId.transfer).family;
            queue_family_count += 1;
        }

        if (queue_usage.present_bit) {
            queue_families[queue_family_count] = queues.get(queues.QueueId.present).family;
            queue_family_count += 1;
        }

        const size_u32 = @intCast(u32, size);

        if (buffer_usage) |usage| {
            const mem_type_index = try handle.findMemoryTypeIndexForBufferInfo(.{
                .flags = .{},
                .size = size_u32,
                .usage = usage,
                .sharing_mode = .exclusive,
                .queue_family_index_count = @intCast(u32, queue_family_count),
                .p_queue_family_indices = &queue_families,
            }, .{
                .flags = .{
                    .withinBudget = true,
                },
                .usage = mem_usage,
            });

            const pool = try handle.createPool(.{
                .memoryTypeIndex = mem_type_index,
                .frameInUseCount = settings.resource_sets - 1,
            });

            return Self{
                .size = size,
                .buffer_usage = usage,
                .mem_usage = mem_usage,
                .queue_usage = queue_usage,
                .handle = pool,
            };
        } else if (image_usage) |usage| {
            const mem_type_index = try handle.findMemoryTypeIndexForImageInfo(.{
                .flags = .{},
                .image_type = .@"2d",
                .format = .r16g16b16a16_sfloat,
                .extent = .{
                    .width = size_u32,
                    .height = size_u32,
                    .depth = 1,
                },
                .mip_levels = 1 + std.math.log2(size_u32),
                .array_layers = 1,
                .samples = .{
                    .@"1_bit" = true,
                },
                .tiling = .optimal,
                .usage = usage,
                .sharing_mode = .exclusive,
                .queue_family_index_count = @intCast(u32, queue_family_count),
                .p_queue_family_indices = &queue_families,
                .initial_layout = .@"undefined",
            }, .{
                .flags = .{
                    .withinBudget = true,
                },
                .usage = mem_usage,
            });

            const pool = try handle.createPool(.{
                .memoryTypeIndex = mem_type_index,
                .frameInUseCount = settings.resource_sets,
            });

            return Self{
                .size = size,
                .image_usage = usage,
                .mem_usage = mem_usage,
                .queue_usage = queue_usage,
                .handle = pool,
            };
        } else {
            return error.InvalidUsageFlags;
        }
    }

    pub fn deinit(self: *const Self, handle: vma.Allocator) void {
        handle.destroyPool(self.handle);
    }
};

pub const Releasable = struct {
    const Self = @This();

    const Type = enum {
        buffer,
        image,
        image_view,
        attachment, // view used as an attachment
    };

    pub const Union = union(Type) {
        buffer: struct {
            handle: vk.Buffer,
            allocation: vma.Allocation,
        },
        image: struct {
            handle: vk.Image,
            allocation: vma.Allocation,
            view: vk.ImageView,
        },
        image_view: vk.ImageView,
        attachment: vk.ImageView,
    };

    frame_count: u64,
    submit_id: SubmitId,
    object: Union,

    const pm_releasableinit = ProfileMark.init("memory.Releasable.init");
    pub fn init(device: *const Device, frame_count: u64, id: SubmitId, object: Union) !Self {
        pm_releasableinit.begin();
        defer pm_releasableinit.end();

        assert(id.valid);
        assert(s_allocator.handle != .Null);

        const self = Self{
            .submit_id = id,
            .object = object,
            .frame_count = frame_count,
        };

        {
            s_allocator.mutex.lock();
            defer s_allocator.mutex.unlock();
            try s_allocator.releasables.append(self);
        }

        try finalizeCheck(device, s_allocator.releasables.items.len);
        return self;
    }

    const pm_releasabledeinit = ProfileMark.init("memory.Releasable.deinit");
    pub fn deinit(self: *const Self, device: *const Device, allocator: Allocator) void {
        pm_releasabledeinit.begin();
        defer pm_releasabledeinit.end();

        switch (self.object) {
            .buffer => |buffer| {
                allocator.handle.destroyBuffer(buffer.handle, buffer.allocation);
            },
            .image => |img| {
                device.dispatch.destroyImageView(device.handle, img.view, null);
                allocator.handle.destroyImage(img.handle, img.allocation);
            },
            .image_view => |image_view| {
                device.dispatch.destroyImageView(device.handle, image_view, null);
            },
            .attachment => |image_view| {
                framebuffer.remove(device, image_view);
                device.dispatch.destroyImageView(device.handle, image_view, null);
            },
        }
    }
};

const Allocator = struct {
    const Self = @This();

    mutex: std.Thread.Mutex,
    handle: vma.Allocator,
    staging_pool: Pool,
    device_buffer_pool: Pool,
    dynamic_buffer_pool: Pool,
    device_texture_pool: Pool,
    releasables: std.ArrayList(Releasable),

    pub fn init(instance: *const Instance, device: *const Device) !Self {
        const funcs = vma.VulkanFunctions.init(instance.handle, device.handle, Base.vk_proc);

        const handle = try vma.Allocator.create(.{
            .vulkanApiVersion = vk.API_VERSION_1_1, // LIES
            .flags = .{
                .memoryBudgetEXT = true,
                .externallySynchronized = true,
            },
            .instance = instance.handle,
            .physicalDevice = device.physical_device,
            .device = device.handle,
            .pVulkanFunctions = &funcs,
            .pAllocationCallbacks = null,
            .frameInUseCount = settings.resource_sets,
        });
        errdefer handle.destroy();

        const staging_pool = try Pool.init(handle, 1 << 20, vk.BufferUsageFlags{
            .transfer_src_bit = true,
            .transfer_dst_bit = true,
        }, null, .cpuOnly, .{
            .transfer_bit = true,
            .graphics_bit = true,
            .compute_bit = true,
        });
        errdefer staging_pool.deinit(handle);

        const device_buffer_pool = try Pool.init(handle, 1 << 20, vk.BufferUsageFlags{
            .transfer_src_bit = true,
            .transfer_dst_bit = true,
            .vertex_buffer_bit = true,
            .index_buffer_bit = true,
            .uniform_buffer_bit = true,
            .storage_buffer_bit = true,
        }, null, .gpuOnly, .{
            .transfer_bit = true,
            .graphics_bit = true,
            .compute_bit = true,
        });
        errdefer device_buffer_pool.deinit(handle);

        const dynamic_buffer_pool = try Pool.init(handle, 1 << 20, vk.BufferUsageFlags{
            .vertex_buffer_bit = true,
            .index_buffer_bit = true,
            .uniform_buffer_bit = true,
        }, null, .cpuToGpu, .{
            .graphics_bit = true,
            .compute_bit = true,
        });
        errdefer dynamic_buffer_pool.deinit(handle);

        const device_texture_pool = try Pool.init(handle, 1024, null, vk.ImageUsageFlags{
            .transfer_dst_bit = true,
            .transfer_src_bit = true,
            .sampled_bit = true,
            .storage_bit = true,
        }, .gpuOnly, .{
            .transfer_bit = true,
            .graphics_bit = true,
            .compute_bit = true,
        });
        errdefer device_texture_pool.deinit(handle);

        return Self{
            .releasables = std.ArrayList(Releasable).init(perm_allocator),
            .mutex = std.Thread.Mutex{},
            .handle = handle,
            .staging_pool = staging_pool,
            .device_buffer_pool = device_buffer_pool,
            .dynamic_buffer_pool = dynamic_buffer_pool,
            .device_texture_pool = device_texture_pool,
        };
    }

    pub fn deinit(self: *Self) void {
        self.staging_pool.deinit(self.handle);
        self.device_texture_pool.deinit(self.handle);
        self.device_buffer_pool.deinit(self.handle);
        self.dynamic_buffer_pool.deinit(self.handle);
        self.handle.destroy();
    }
};

const FinalizeCheckError = Command.Buffer.FlushQueueTransfersError || SubmitId.WaitAllError || error{FinalizeUpdateFail};
fn finalizeCheck(device: *const Device, size: usize) FinalizeCheckError!void {
    if (size >= 1024) { // TODO convar
        std.debug.print("Too many gpu objects, force-finalizing\n", .{});
        try Command.flush(device);
        try SubmitId.waitAll(device);
        update(device) catch return error.FinalizeUpdateFail;
    }
}
