const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");

const vec = @import("vec");
const Vec2f = vec.Vec2f;
const Vec3f = vec.Vec3f;
const Vec4f = vec.Vec4f;

const Vec2i = vec.Vec2i;
const Vec3i = vec.Vec3i;
const Vec4i = vec.Vec4i;

const Device = @import("device.zig").Device;
const FixedFuncs = @import("Pass.zig").FixedFuncs;
const VertexLayout = @import("Pass.zig").VertexLayout;

pub const Layout = struct {
    pub fn init(device: *const Device, set_layouts: []const vk.DescriptorSetLayout, ranges: []const vk.PushConstantRange) !vk.PipelineLayout {
        return try device.dispatch.createPipelineLayout(device.handle, &.{
            .flags = .{},
            .set_layout_count = @intCast(u32, set_layouts.len),
            .p_set_layouts = set_layouts.ptr,
            .push_constant_range_count = @intCast(u32, ranges.len),
            .p_push_constant_ranges = ranges.ptr,
        }, null);
    }

    pub fn deinit(device: *const Device, layout: vk.PipelineLayout) void {
        device.dispatch.destroyPipelineLayout(device.handle, layout, null);
    }
};

pub const VertType = enum {
    float,
    float2,
    float3,
    float4,
    int,
    int2,
    int3,
    int4,

    pub fn size(t: VertType) i32 {
        return switch (t) {
            .float => @sizeOf(f32),
            .float2 => @sizeOf(Vec2f),
            .float3 => @sizeOf(Vec3f),
            .float4 => @sizeOf(Vec4f),
            .int => @sizeOf(i32),
            .int2 => @sizeOf(Vec2i),
            .int3 => @sizeOf(Vec3i),
            .int4 => @sizeOf(Vec4i),
            else => assert(false),
        };
    }

    pub fn format(t: VertType) vk.Format {
        return switch (t) {
            .float => .r32_sfloat,
            .float2 => .r32g32_sfloat,
            .float3 => .r32g32b32_sfloat,
            .float4 => .r32g32b32a32_sfloat,
            .int => .r32_sint,
            .int2 => .r32g32_sint,
            .int3 => .r32g32b32_sint,
            .int4 => .r32g32b32a32_sint,
            else => assert(false),
        };
    }
};

pub fn newGfx(device: *const Device, fixed_funcs: *const FixedFuncs, vert_layout: *const VertexLayout, layout: vk.PipelineLayout, render_pass: vk.RenderPass, subpass: ?u32, shaders: []const vk.PipelineShaderStageCreateInfo) !vk.Pipeline {
    assert(layout != .null_handle);
    assert(render_pass != .null_handle);

    const vertex_input_info = vk.PipelineVertexInputStateCreateInfo{
        .flags = .{},
        .vertex_binding_description_count = @intCast(u32, vert_layout.bindings.len),
        .p_vertex_binding_descriptions = vert_layout.bindings.ptr,
        .vertex_attribute_description_count = @intCast(u32, vert_layout.attributes.len),
        .p_vertex_attribute_descriptions = vert_layout.attributes.ptr,
    };

    const input_assembly = vk.PipelineInputAssemblyStateCreateInfo{
        .flags = .{},
        .primitive_restart_enable = 0,
        .topology = fixed_funcs.topology,
    };

    const viewport_state = vk.PipelineViewportStateCreateInfo{
        .flags = .{},
        .viewport_count = 1,
        .p_viewports = @ptrCast([*]const vk.Viewport, &fixed_funcs.viewport),
        .scissor_count = 1,
        .p_scissors = @ptrCast([*]const vk.Rect2D, &fixed_funcs.scissor),
    };

    const rasterizer = vk.PipelineRasterizationStateCreateInfo{
        .flags = .{},
        .depth_clamp_enable = @boolToInt(fixed_funcs.depth_clamp),
        .rasterizer_discard_enable = 0,
        .depth_bias_constant_factor = 0,
        .depth_bias_enable = 0,
        .depth_bias_clamp = 0,
        .depth_bias_slope_factor = 0,
        .polygon_mode = fixed_funcs.polygon_mode,
        .cull_mode = fixed_funcs.cull_mode,
        .front_face = fixed_funcs.front_face,
        .line_width = 1,
    };

    const multisampling = vk.PipelineMultisampleStateCreateInfo{
        .flags = .{},
        .sample_shading_enable = 0,
        .min_sample_shading = 0,
        .p_sample_mask = null,
        .alpha_to_coverage_enable = 0,
        .alpha_to_one_enable = 0,
        .rasterization_samples = .{ .@"1_bit" = true },
    };

    const depth_stencil = vk.PipelineDepthStencilStateCreateInfo{
        .flags = .{},
        .depth_test_enable = @boolToInt(fixed_funcs.depth_test_enable),
        .depth_bounds_test_enable = 0,
        .stencil_test_enable = 0,
        .front = std.mem.zeroes(vk.StencilOpState),
        .back = std.mem.zeroes(vk.StencilOpState),
        .depth_write_enable = @boolToInt(fixed_funcs.depth_write_enable),
        .depth_compare_op = fixed_funcs.depth_compare_op,
        .min_depth_bounds = 0,
        .max_depth_bounds = 1,
    };

    var color_blend_attachments = std.mem.zeroes([8]vk.PipelineColorBlendAttachmentState);
    if (fixed_funcs.attachments) |attachments| {
        for (attachments) |attachment, i| {
            color_blend_attachments[i] = vk.PipelineColorBlendAttachmentState{
                .blend_enable = @boolToInt(attachment.blend_enable),
                .src_color_blend_factor = attachment.src_color_blend_factor,
                .dst_color_blend_factor = attachment.dst_color_blend_factor,
                .color_blend_op = attachment.color_blend_op,
                .src_alpha_blend_factor = attachment.src_alpha_blend_factor,
                .dst_alpha_blend_factor = attachment.dst_alpha_blend_factor,
                .alpha_blend_op = attachment.alpha_blend_op,
                .color_write_mask = attachment.color_write_mask,
            };
        }
    }

    const color_blending = vk.PipelineColorBlendStateCreateInfo{
        .flags = .{},
        .logic_op_enable = 0,
        .logic_op = .clear,
        .blend_constants = std.mem.zeroes([4]f32),
        .attachment_count = if (fixed_funcs.attachments) |att| @intCast(u32, att.len) else 0,
        .p_attachments = &color_blend_attachments,
    };

    const dynamic_states = [_]vk.DynamicState{
        .viewport,
        .scissor,
    };

    const dynamic_state_info = vk.PipelineDynamicStateCreateInfo{
        .flags = .{},
        .dynamic_state_count = dynamic_states.len,
        .p_dynamic_states = &dynamic_states,
    };

    const pipeline_info = vk.GraphicsPipelineCreateInfo{
        .flags = .{},
        .stage_count = @intCast(u32, shaders.len),
        .p_stages = shaders.ptr,
        .p_vertex_input_state = &vertex_input_info,
        .p_input_assembly_state = &input_assembly,
        .p_tessellation_state = null,
        .p_viewport_state = &viewport_state,
        .p_rasterization_state = &rasterizer,
        .p_multisample_state = &multisampling,
        .p_depth_stencil_state = &depth_stencil,
        .p_color_blend_state = &color_blending,
        .p_dynamic_state = &dynamic_state_info,
        .layout = layout,
        .render_pass = render_pass,
        .base_pipeline_handle = .null_handle,
        .base_pipeline_index = 0,
        .subpass = subpass orelse 0,
    };

    var pipeline: vk.Pipeline = undefined;
    _ = try device.dispatch.createGraphicsPipelines(device.handle, .null_handle, 1, @ptrCast([*]const vk.GraphicsPipelineCreateInfo, &pipeline_info), null, @ptrCast([*]vk.Pipeline, &pipeline));
    return pipeline;
}

pub fn newComp(device: *const Device, layout: vk.PipelineLayout, shader: *const vk.PipelineShaderStageCreateInfo) !vk.Pipeline {
    assert(layout != .null_handle);

    var pipelines: [*]vk.Pipeline = undefined;
    _ = try device.dispatch.createComputePipelines(device.handle, .null_handle, 1, @ptrCast([*]const vk.ComputePipelineCreateInfo, &.{
        .flags = .{},
        .stage = shader,
        .layout = layout,
    }), null, pipelines);

    return pipelines[0];
}

pub fn deinit(device: *const Device, pipeline: vk.Pipeline) void {
    device.dispatch.destroyPipeline(device.handle, pipeline, null);
}
