const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");

const Device = @import("device.zig").Device;
const DeviceDispatch = @import("device.zig").DeviceDispatch;
const queues = @import("queues.zig");
const profiler = @import("../../common/profiler.zig");
const Fence = @import("sync.zig").Fence;

pub const SubmitId = packed struct {
    counter: u32 = 0,
    queue_id: queues.QueueId = .graphics,
    valid: bool = false,

    pub const PollError = Fence.StatError || DeviceDispatch.ResetCommandBufferError;
    const pm_poll = profiler.ProfileMark.init("SubmitId.poll");
    pub fn poll(self: *const SubmitId, device: *const Device) PollError!bool {
        pm_poll.begin();
        defer pm_poll.end();

        assert(self.valid);
        const queue = queues.get(self.queue_id);

        const ring_mask = queue.cmds.len - 1;
        while (self.counter -% queue.tail < queue.cmds.len) : (queue.tail += 1) {
            const i = queue.tail & ring_mask;
            if ((try queue.cmd_fences[i].stat(device)) != .signaled) {
                return false;
            }

            try device.dispatch.resetCommandBuffer(queue.cmds[i], .{});
            queue.cmd_ids[i] = 0;
        }

        return true;
    }

    const pm_wait = profiler.ProfileMark.init("SubmitId.wait");
    pub fn wait(self: *const SubmitId, device: *const Device) !void {
        pm_wait.begin();
        defer pm_wait.end();

        std.debug.assert(self.valid);

        const queue = queues.get(self.queue_id);

        const ring_mask = queue.cmds.len - 1;

        while ((self.counter -% queue.tail) < queue.cmds.len) : (queue.tail += 1) {
            const i = queue.tail & ring_mask;
            std.debug.assert(queue.cmd_ids[i] == queue.tail);
            try queue.cmd_fences[i].wait(device);
            device.dispatch.resetCommandBuffer(queue.cmds[i], .{}) catch {};
            queue.cmd_ids[i] = 0;
        }
    }

    pub const WaitAllError = Fence.WaitError || DeviceDispatch.ResetCommandBufferError;
    const pm_waitall = profiler.ProfileMark.init("SubmitId.waitAll");
    pub fn waitAll(device: *const Device) !void {
        pm_waitall.begin();
        defer pm_waitall.end();

        var queueId: usize = 0;
        while (queueId < queues.QueueId.count) : (queueId += 1) {
            const queue = queues.get(@intToEnum(queues.QueueId, queueId));
            const ring_mask = queue.cmds.len - 1;
            while (queue.tail != queue.head) : (queue.tail += 1) {
                const i = queue.tail & ring_mask;
                assert(queue.cmd_ids[i] == queue.tail);
                try queue.cmd_fences[i].wait(device);
                try device.dispatch.resetCommandBuffer(queue.cmds[i], .{});
                queue.cmd_ids[i] = 0;
            }
        }
    }
};
