const std = @import("std");
const assert = std.debug.assert;
const vk = @import("vulkan");
const math = @import("std").math;
const Command = @import("Command.zig");
const profiler = @import("../../common/profiler.zig");
const settings = @import("settings.zig");
const Instance = @import("instance.zig").Instance;
const Device = @import("device.zig").Device;
const Buffer = @import("Buffer.zig");
const Window = @import("display.zig").Window;
const sync = @import("sync.zig");
const queues = @import("queues.zig");
const Image = @import("image.zig").Image;
const SubmitId = @import("submit_id.zig").SubmitId;

// TODO memory
var gpa = std.heap.GeneralPurposeAllocator(.{}){};
const allocator = gpa.allocator();

const preferred_present_modes = [_]vk.PresentModeKHR{
    .mailbox_khr,
    .immediate_khr,
    .fifo_relaxed_khr,
    .fifo_khr,
};

const preferred_surface_formats = [_]vk.SurfaceFormatKHR{
    // PQ Rec2100
    // https://en.wikipedia.org/wiki/Rec._2100
    // https://en.wikipedia.org/wiki/High-dynamic-range_video#Perceptual_quantizer
    if (settings.hdr_on) .{ .format = .a2r10g10b10_unorm_pack32, .color_space = .hdr10_st2084_ext } else void,
    if (settings.hdr_on) .{ .format = .a2b10g10r10_unorm_pack32, .color_space = .hdr10_st2084_ext } else void,
    // 10-bit sRGB
    if (settings.hdr_on) .{ .format = .a2r10g10b10_unorm_pack32, .color_space = .srgb_nonlinear_khr } else void,
    if (settings.hdr_on) .{ .format = .a2b10g10r10_unorm_pack32, .color_space = .srgb_nonlinear_khr } else void,
    // 8 bit sRGB
    .{ .format = .r8g8b8a8_srgb, .color_space = .srgb_nonlinear_khr },
    .{ .format = .b8g8r8a8_srgb, .color_space = .srgb_nonlinear_khr },
};

const Support = struct {
    const Self = @This();

    capabilities: vk.SurfaceCapabilitiesKHR,
    formats: []vk.SurfaceFormatKHR,
    modes: []vk.PresentModeKHR,

    pub fn init(instance: *const Instance, device: *const Device, window: *const Window) !Self {
        const capabilities = try instance.dispatch.getPhysicalDeviceSurfaceCapabilitiesKHR(device.physical_device, window.surface);

        var format_count: u32 = undefined;
        _ = try instance.dispatch.getPhysicalDeviceSurfaceFormatsKHR(device.physical_device, window.surface, &format_count, null);
        const formats = try allocator.alloc(vk.SurfaceFormatKHR, format_count);
        errdefer allocator.free(formats);
        _ = try instance.dispatch.getPhysicalDeviceSurfaceFormatsKHR(device.physical_device, window.surface, &format_count, formats.ptr);

        var mode_count: u32 = undefined;
        _ = try instance.dispatch.getPhysicalDeviceSurfacePresentModesKHR(device.physical_device, window.surface, &mode_count, null);
        const modes = try allocator.alloc(vk.PresentModeKHR, format_count);
        errdefer allocator.free(modes);
        _ = try instance.dispatch.getPhysicalDeviceSurfacePresentModesKHR(device.physical_device, window.surface, &mode_count, modes.ptr);

        return Self{
            .capabilities = capabilities,
            .formats = formats,
            .modes = modes,
        };
    }

    pub fn deinit(self: *const Self) void {
        allocator.free(self.formats);
        allocator.free(self.modes);
    }

    pub fn selectSwapFormat(self: *const Self) !vk.SurfaceFormatKHR {
        if (self.formats.len <= 0) {
            return error.NoSurfaceFormats;
        }

        for (preferred_surface_formats) |preferred_format| {
            for (self.formats) |format| {
                if (std.meta.eql(format, preferred_format)) {
                    return format;
                }
            }
        }

        return self.formats[0];
    }

    pub fn selectSwapMode(self: *const Self) !vk.PresentModeKHR {
        if (self.modes.len <= 0) {
            return error.NoPresentModes;
        }

        for (preferred_present_modes) |preferred_mode| {
            for (self.modes) |mode| {
                if (preferred_mode == mode) {
                    return mode;
                }
            }
        }

        return self.modes[0];
    }

    pub fn selectSwapExtent(self: *const Self, window: *const Window) vk.Extent2D {
        if (self.capabilities.current_extent.width != ~@as(u32, 0)) {
            return self.capabilities.current_extent;
        }

        const min_ext = self.capabilities.min_image_extent;
        const max_ext = self.capabilities.max_image_extent;
        return vk.Extent2D{
            .width = math.clamp(window.size.width, min_ext.width, max_ext.width),
            .height = math.clamp(window.size.height, min_ext.height, max_ext.height),
        };
    }
};

pub const Swapchain = struct {
    const Self = @This();

    handle: vk.SwapchainKHR,
    color_format: vk.Format,
    color_space: vk.ColorSpaceKHR,
    mode: vk.PresentModeKHR,
    width: u32,
    height: u32,

    length: u32,
    image_index: usize = 0,
    image_submits: [settings.max_swapchain_len]SubmitId = std.mem.zeroes([settings.max_swapchain_len]SubmitId),
    images: [settings.max_swapchain_len]Image = std.mem.zeroes([settings.max_swapchain_len]Image),

    sync_index: usize = 0,
    sync_submits: [settings.resource_sets]SubmitId = std.mem.zeroes([settings.resource_sets]SubmitId),
    available_semas: [settings.resource_sets]sync.Semaphore = std.mem.zeroes([settings.resource_sets]sync.Semaphore),
    rendered_semas: [settings.resource_sets]sync.Semaphore = std.mem.zeroes([settings.resource_sets]sync.Semaphore),

    pub fn init(instance: *const Instance, window: *const Window, device: *const Device, previous: ?*Self) !Self {
        const support = try Support.init(instance, device, window);
        defer support.deinit();

        const queue_support = try queues.Support.init(instance, device.physical_device, window);
        defer queue_support.deinit();

        const format = try support.selectSwapFormat();
        const mode = try support.selectSwapMode();
        const ext = support.selectSwapExtent(window);

        var img_count = math.clamp(settings.desired_swapchain_len, support.capabilities.min_image_count, math.min(settings.max_swapchain_len, support.capabilities.max_image_count));

        const families = [_]u32{
            queue_support.families[@enumToInt(queues.QueueId.graphics)] orelse unreachable,
            queue_support.families[@enumToInt(queues.QueueId.present)] orelse unreachable,
        };
        const concurrent = families[0] != families[1];

        const usage: vk.ImageUsageFlags = .{ .color_attachment_bit = true };

        const swap_info = vk.SwapchainCreateInfoKHR{
            .flags = .{},
            .surface = window.surface,
            .present_mode = mode,
            .min_image_count = img_count,
            .image_format = format.format,
            .image_color_space = format.color_space,
            .image_extent = ext,
            .image_array_layers = 1,
            .image_usage = usage,
            .image_sharing_mode = if (concurrent) .concurrent else .exclusive,
            .queue_family_index_count = if (concurrent) families.len else 0,
            .p_queue_family_indices = &families,
            .pre_transform = support.capabilities.current_transform,
            // no compositing with window manager / desktop background
            .composite_alpha = .{ .opaque_bit_khr = true },
            // don't render pixels behind other windows
            .clipped = 1,
            // prev swapchain, if recreating
            .old_swapchain = if (previous) |p| p.handle else .null_handle,
        };

        const handle = try device.dispatch.createSwapchainKHR(device.handle, &swap_info, null);
        errdefer device.dispatch.destroySwapchainKHR(device.handle, handle, null);

        if (previous == null) {
            std.debug.print("Present mode: '{s}'\n", .{@tagName(mode)});
            std.debug.print("Present extent: {} x {}\n", .{ ext.width, ext.height });
            std.debug.print("Present images: {}\n", .{img_count});
            std.debug.print("Present sharing mode: '{s}'\n", .{if (concurrent) @as([]const u8, "Concurrent") else "Exclusive"});
            std.debug.print("Color space: '{s}'\n", .{@tagName(format.color_space)});
            std.debug.print("Format: '{s}'\n", .{@tagName(format.format)});
        }

        var images: [settings.max_swapchain_len]vk.Image = std.mem.zeroes([settings.max_swapchain_len]vk.Image);
        _ = try device.dispatch.getSwapchainImagesKHR(device.handle, handle, &img_count, null);
        if (img_count > settings.max_swapchain_len) {
            return error.TooManyImages;
        }

        _ = try device.dispatch.getSwapchainImagesKHR(device.handle, handle, &img_count, &images);

        var self = Self{
            .handle = handle,
            .mode = mode,
            .color_format = format.format,
            .color_space = format.color_space,
            .width = ext.width,
            .height = ext.height,
            .length = img_count,
        };

        for (images[0..img_count]) |img, i| {
            try self.images[i].import(device, .{
                .flags = .{},
                .image_type = .@"2d",
                .extent = .{
                    .width = ext.width,
                    .height = ext.height,
                    .depth = 1,
                },
                .format = format.format,
                .mip_levels = 1,
                .array_layers = 1,
                .samples = .{ .@"1_bit" = true },
                .tiling = .optimal,
                .usage = usage,
                .sharing_mode = swap_info.image_sharing_mode,
                .queue_family_index_count = swap_info.queue_family_index_count,
                .p_queue_family_indices = swap_info.p_queue_family_indices,
                .initial_layout = .@"undefined",
            }, img);
        }

        {
            var i: usize = 0;
            while (i < settings.resource_sets) : (i += 1) {
                self.available_semas[i] = try sync.Semaphore.init(device);
                self.rendered_semas[i] = try sync.Semaphore.init(device);
            }
        }

        return self;
    }

    pub fn deinit(self: *const Self, device: *const Device) void {
        device.waitIdle() catch {};

        for (self.available_semas) |_, i| {
            self.available_semas[i].deinit(device);
            self.rendered_semas[i].deinit(device);
        }

        for (self.images) |image| {
            image.deinit(device);
        }
    }

    const pm_acquiresync = profiler.ProfileMark.init("swapchain.Swapchain.acquireSync");
    pub fn acquireSync(self: *Self, device: *const Device) !usize {
        pm_acquiresync.begin();
        defer pm_acquiresync.end();

        const sync_index = (self.sync_index + 1) % self.sync_submits.len;
        const sub = self.sync_submits[sync_index];
        if (sub.valid) {
            try sub.wait(device);
        }

        self.sync_index = sync_index;
        return sync_index;
    }

    const pm_acquireimage = profiler.ProfileMark.init("swapchain.Swapchain.acquireImage");
    pub fn acquireImage(self: *Self, device: *const Device) !usize {
        pm_acquireimage.begin();
        defer pm_acquireimage.end();

        const timeout = std.math.maxInt(u64);
        const result = try device.dispatch.acquireNextImageKHR(device.handle, self.handle, timeout, self.available_semas[self.sync_index].handle, .null_handle);
        assert(result.image_index < self.length);
        const sub = self.image_submits[result.image_index];
        if (sub.valid) {
            try sub.wait(device);
        }

        self.image_index = result.image_index;
        return result.image_index;
    }

    const pm_submit = profiler.ProfileMark.init("swapchain.Swapchain.submit");
    pub fn submit(self: *Self, command: *Command.Buffer, device: *const Device) !void {
        pm_submit.begin();
        defer pm_submit.end();

        const image_index = self.image_index;
        const sync_index = self.sync_index;
        {
            var backbuf = self.getBackBuffer();
            const prev_use = backbuf.state.stage;
            _ = try Image.State.presentSrc(device, command, backbuf);

            const submit_id = try command.submit(device, self.available_semas[sync_index], prev_use, self.rendered_semas[sync_index]);
            self.sync_submits[sync_index] = submit_id;
            self.image_submits[image_index] = submit_id;
        }

        {
            var queue = queues.get(.graphics);
            if (!queue.pres) {
                queue = queues.get(.present);
            }

            _ = try device.dispatch.queuePresentKHR(queue.handle, &.{
                .wait_semaphore_count = 1,
                .p_wait_semaphores = @ptrCast([*]const vk.Semaphore, &self.rendered_semas[sync_index]),
                .swapchain_count = 1,
                .p_swapchains = @ptrCast([*]const vk.SwapchainKHR, &self.handle),
                .p_results = null,
                .p_image_indices = @ptrCast([*]const u32, &self.image_index),
            });
        }
    }

    pub fn getViewport(self: *const Self) vk.Viewport {
        assert(self.handle != .null_handle);
        return vk.Viewport{
            .x = 0.0,
            .y = 0.0,
            .width = @intToFloat(f32, self.width),
            .height = @intToFloat(f32, self.height),
            .min_depth = 0.0,
            .max_depth = 1.0,
        };
    }

    pub fn getRect(self: *const Self) vk.Rect2D {
        assert(self.handle != .null_handle);
        return vk.Rect2D{
            .extent = .{
                .width = self.width,
                .height = self.height,
            },
            .offset = std.mem.zeroes(vk.Offset2D),
        };
    }

    pub fn getAspect(self: *const Self) f32 {
        assert(self.handle != .null_handle);
        return @intToFloat(f32, self.width) / @intToFloat(f32, self.height);
    }

    pub fn getSwapIndex(self: *const Self) usize {
        return self.image_index;
    }

    pub fn getBackBufferRO(self: *const Self) *const Image {
        const img = &self.images[self.getSwapIndex()];
        assert(img.handle != .null_handle);
        return img;
    }

    pub fn getBackBuffer(self: *Self) *Image {
        var img = &self.images[self.getSwapIndex()];
        assert(img.handle != .null_handle);
        return img;
    }

    pub fn getSyncIndex(self: *const Swapchain) usize {
        return self.sync_index;
    }

    pub fn getNextSyncIndex(self: *const Swapchain) usize {
        return (self.getSyncIndex() + 1) % settings.resource_sets;
    }
    pub fn getPrevSyncIndex(self: *const Swapchain) usize {
        return (self.getSyncIndex() + (settings.resource_sets - 1)) % settings.resource_sets;
    }
};
